from abc import abstractmethod
import sys, os, pathlib, re, abc, contextlib, collections, time, json, inspect
import itertools
from collections import namedtuple
from collections.abc import Mapping
from pathlib import Path

from ..core.rule import IRule
from .rule import Rule
from .igroup import IGroup
from .file import File, VFile, IFile, IVFile
from . import events as group_events
from .event_logger import create_event_callback
from . import graphviz
from ..core.make import make as _make, make_multi_thread, Event
from ..logwriter.writer import \
    TextWriter, ColorTextWriter, HTMLJupyterWriter, \
    term_is_jupyter, TextFileWriterOpenOnDemand, HTMLFileWriterOpenOnDemand

from ..utils.nest import \
    StructKey, map_structure, flatten, struct_get, \
    flatten_to_struct_keys, pack_sequence_as



class IFileNode:
    @property
    @abstractmethod
    def path(self): ...

    @abstractmethod
    def touch(self, _t): ...

    @abstractmethod
    def clean(self): ...


class FileNodeAtom(IFileNode):
    def __init__(self, tree_info, file):
        assert isinstance(file, IFile)
        self._info = tree_info
        self._file = file

    @property
    def path(self):
        return self._file.path

    @property
    def abspath(self):
        return self._file.abspath

    def touch(self, _t=None):
        if _t is None:
            _t = time.time()
        open(self._file.path, 'w').close()
        os.utime(self._file.path, (_t, _t))
        self._info.callback(group_events.Touch(self.path))

    def clean(self):
        try:
            os.remove(self._file.path)
            self._info.callback(group_events.Clean(self.path))
        except:
            pass

    def __repr__(self):
        return f'FileNodeAtom("{self.path}")'


class FileNodeTuple(tuple, IFileNode):
    def __new__(cls, lst):
        return super().__new__(cls, lst)

    def __init__(self, lst): ...

    @property
    def path(self):
        return tuple(x.path for x in self)

    @property
    def abspath(self):
        return tuple(x.abspath for x in self)

    def touch(self, _t=None):
        if _t is None:
            _t = time.time()
        for x in self: x.touch(_t)

    def clean(self):
        for x in self: x.clean()

    def __repr__(self):
        return f'FileNodeTuple{super().__repr__()}'


class FileNodeDict(Mapping, IFileNode):
    def __init__(self, dic):
        self._dic = dic

    @property
    def path(self):
        return {k: v.path for k,v in self._dic.items()}

    @property
    def abspath(self):
        return {k: v.abspath for k,v in self._dic.items()}

    def touch(self, _t):
        if _t is None: _t = time.time()
        for k,v in self._dic.items(): v.touch(_t)

    def clean(self):
        for k,v in self._dic.items(): v.clean()

    def __getitem__(self, k):
        return self._dic[k]

    def __iter__(self):
        return self._dic.__iter__()

    def __len__(self):
        return len(self._dic)

    def __eq__(self, other):
        return self is other

    def __hash__(self):
        return hash(id(self))

    def __repr__(self):
        return f'FileNodeDict{dict(self)}'


class RuleNodeBase:
    def __init__(self, name, rule, group_tree_info):
        self._rule = rule
        self._info = group_tree_info
        self._name = name

    def make(
        self,
        dry_run=False,
        keep_going=False,
        *,
        nthreads=1
    ):
        make(
            self,
            dry_run=dry_run,
            keep_going=keep_going,
            nthreads=nthreads
        )

    @property
    def name(self):
        return self._name


class RuleNodeAtom(RuleNodeBase, FileNodeAtom):
    def __init__(self, name, rule, group_tree_info, file):
        RuleNodeBase.__init__(self, name, rule, group_tree_info)
        FileNodeAtom.__init__(self, group_tree_info, file)


    def __repr__(self):
        return f'RuleNodeAtom(path="{self.path}")'


class RuleNodeTuple(RuleNodeBase, FileNodeTuple):
    def __new__(cls, _name, _rule, _group_tree_info, lst):
        return FileNodeTuple.__new__(cls, lst)

    def __init__(self, name, rule, group_tree_info, lst):
        RuleNodeBase.__init__(self, name, rule, group_tree_info)
        FileNodeTuple.__init__(self, lst)


    def __repr__(self):
        return f'RuleNodeTuple{tuple(self)}'


class RuleNodeDict(RuleNodeBase, FileNodeDict):
    def __new__(cls, _name, _rule, _group_tree_info, dic):
        return FileNodeDict.__new__(cls)

    def __init__(self, name, rule, group_tree_info, dic):
        RuleNodeBase.__init__(self, name, rule, group_tree_info)
        FileNodeDict.__init__(self, dic)
        
    def __repr__(self):
        return f'RuleNodeDict{dict(self)}'


class GroupTreeInfo:
    def __init__(self, logwriters):
        self.path_to_rule = {}
        self.path_to_file = {}
        self.rule_to_name = {}

        self.callback = create_event_callback(logwriters, self.rule_to_name)

    
class Group(IGroup):
    def __init__(self, info, prefix, name):
        if not isinstance(prefix, str):
            raise TypeError('prefix must be str')

        self._info = info
        self._prefix = prefix
        self._name = name
        self._children = {}


    def add_group(self, name, dirname = None, *, prefix = None):
        """
        Call signatures:
            add_group(name, [dirname])
            add_group(name, prefix=prefix)
        """
        if isinstance(name, os.PathLike):
            name = name.__fspath__()
        elif not isinstance(name, str):
            raise TypeError('name must be str or os.PathLike')

        if name in self._children:
            raise KeyError(f'name {repr(name)} already exists in this Group')

        if not _is_valid_node_name(name):
            raise ValueError(f'name "{name}" contains some illegal characters')

        if name == '':
            raise ValueError(f'name must not be ""')

        if dirname is not None and prefix is not None:
            raise TypeError('Either dirname or prefix can be specified')

        if dirname is None and prefix is None:
            dirname = name

        if dirname is not None:
            assert isinstance(dirname, (str, os.PathLike))
            prefix = str(Path(dirname)) + os.path.sep

        assert isinstance(prefix, (str, os.PathLike))

        if isinstance(prefix, os.PathLike):
            prefix = prefix.__fspath__()

        if not os.path.isabs(prefix):
            prefix = self._prefix + prefix

        g = Group(self._info, prefix, (*self._name, name))
        
        self._children[name] = g

        if _ismembername(name):
            self.__dict__[name] = g

        return g


    def make(
        self,
        dry_run=False,
        keep_going=False,
        *,
        nthreads=1
    ):
        make(
            self,
            dry_run=dry_run,
            keep_going=keep_going,
            nthreads=nthreads
        )


    # APIs
    def add(self, name, *args, force_update=False, **kwargs):
        """
        Call signatures:
            add(name, [path], method, *args, force_update=False, **kwargs)
            add(name, [path], None, *args, force_update=False, **kwargs)
        """
        if not isinstance(name, str):
            raise ValueError(f'name must be str')

        if len(args) == 0:
            raise TypeError('method must be specified')

        if callable(args[0]) or args[0] is None:
            path = str(name)
            method, *args = args
        else:
            if not (len(args) >= 2 and (callable(args[1]) or args[1] is None)):
                raise TypeError('method must be specified')

            path, method, *args = args

        if method is None:
            def adder(method):
                return self._add(name, path, method, *args, **kwargs)
            
            return adder

        return self._add(
            name, path, method, *args, force_update=force_update, **kwargs)


    def addvf(self, name, *args, force_update=False, **kwargs):
        """Append a VFile (Value File) as a child of this group
        Call signatures:
            addvf(name, method, *args, force_update=False, **kwargs)
            addvf(name, path, method, *args, force_update=False, **kwargs)
        """
        if not isinstance(name, str):
            raise ValueError(f'name must be str')

        if len(args) == 0:
            raise TypeError('method must be specified')

        if callable(args[0]) or args[0] is None:
            path = str(name)
            method, *args = args
        else:
            if not (len(args) >= 2 and (callable(args[1]) or args[1] is None)):
                raise TypeError('method must be specified')

            path, method, *args = args

        if method is None:
            def adder(method):
                assert callable(method)
                return self.add_vf(name, path, method, *args, **kwargs)
            
            return adder

        def wrap_by_VFile(p):
            if isinstance(p, (str, os.PathLike)):
                return VFile(p)
            else:
                return p

        path = map_structure(wrap_by_VFile, path)
            
        return self._add(
            name, path, method, *args, force_update=force_update, **kwargs)


    def _add(
        self, name, files,
        method, *args, force_update=False, **kwargs
    ):
        assert isinstance(name, str)
        assert callable(method)

        path_to_rule = self._info.path_to_rule
        path_to_file = self._info.path_to_file

        if not _is_valid_node_name(name):
            raise ValueError(f'name "{name}" contains some illegal characters')
        if name in self._children:
            raise KeyError(f'name `{name}` already exists')

        if name == '':
            raise ValueError(f'name must not be ""')

        # wrap str/os.PathLike in yfiles by File
        def wrap_by_File(p):
            if isinstance(p, (str, os.PathLike)):
                return File(p)
            else:
                assert isinstance(p, IFile)
                return p

        files = map_structure(wrap_by_File, files)

        # add prefix to paths of yfiles if necessary 
        def add_pfx(f):
            if os.path.isabs(f.path):
                return f
            else:
                return f.__class__(self._prefix + str(f.path))
                # TODO: __init__ of f's class may take args other than path

        files = map_structure(add_pfx, files)

        # expand SELFs in args
        _expanded = False
        def expand_self(arg):
            nonlocal _expanded
            if isinstance(arg, StructKey):
                _expanded = True
                try:
                    return struct_get(files, arg)
                except:
                    raise ValueError(f'Invalid keys for SELF')
            else:
                return arg

        args, kwargs = map_structure(expand_self, (args, kwargs))

        if not _expanded:
            args = (files, *args)

        # validate method signature
        try:
            inspect.signature(method).bind(*args, **kwargs)
        except Exception as e:
            raise Exception(f'Method signature and args/kwargs do not match') from e

        # flatten yfiles and args (for convenience)
        try:
            files_ = flatten(files)
        except Exception as e:
            raise Exception(
                f'Failed to flatten the input file structure. '
                f'This error occurs when the structure contains a dict '
                f'whose keys are not sortable.'
            ) from e

        try:
            args_ = flatten((args, kwargs))
        except Exception as e:
            raise Exception(
                f'Failed to flatten the structure of the args/kwargs. '
                f'This error occurs when it contain a dict '
                f'whose keys are not sortable.'
            ) from e
        
        if len(files_) == 0:
            raise ValueError('at least 1 output file must be specified')
                
        # Unwrap FileNodeAtom
        args_ = [x._file if isinstance(x, FileNodeAtom) else x for x in args_]

        # normalize paths (take the shorter one of absolute or relative path)
        def _shorter_path(p):
            cwd = os.getcwd()
            try:
                rel = Path(os.path.relpath(p, cwd))
                return rel if len(str(rel)) < len(str(p)) else p
            except:
                return p

        _norm = lambda f: \
            f.__class__(_shorter_path(f.path)) if isinstance(f, IFile) else f
        files_ = list(map(_norm, files_))
        args_ = list(map(_norm, args_))

        # check IFile consistency
        p2f = {}
        for x in args_:
            if isinstance(x, IFile):
                if \
                    (x.path in p2f and p2f[x.path] != x) or \
                    (
                        x.abspath in path_to_file and
                        path_to_file[x.abspath] != x
                    ):
                    raise TypeError(
                        f'Inconsistency in IFiles of path {x.path}: '
                        f'One is {x} and the other is {p2f[x.path]}'
                    )
                else:
                    p2f[x.path] = x

        # check duplicate registration of output files
        for f in files_:
            if f.abspath in path_to_rule:
                raise ValueError(
                    f'path {f.path} is already used by another rule: '
                    f'{path_to_rule[f.abspath]}'
                )

        # check if all the y files are included in the arguments
        unused = set(files_) - {f for f in args_ if isinstance(f, IFile)}
        if len(unused) != 0:
            raise ValueError(
                f'Some files are not passed to the method: '
                f'{list(unused)}'
            )
        
        # create deplist
        deplist = []
        _added = set()
        for f in args_:
            if isinstance(f, IFile) and path_to_rule.get(f.abspath) is not None:
                if path_to_rule[f.abspath] not in _added:
                    deplist.append(path_to_rule[f.abspath])
                    _added.add(path_to_rule[f.abspath])

        # create xfiles
        ypaths = set(f.path for f in files_)
        try:
            arg_keys = flatten_to_struct_keys(args)
        except Exception as e:
            raise Exception(
                f'Failed to flatten keyword arguments. '
                f'This error occurs when args/kwargs contain a dict '
                f'whose keys are not sortable.'
            ) from e

        xfiles = [
            (k,f) for k,f in zip(arg_keys, args_)
            if isinstance(f, IFile) and f.path not in ypaths
        ]

        # check if keys for IVFiles are JSON convertible
        def _assert_key_json_convertible(key, f):
            try:
                json.dumps(key)
            except Exception as e:
                raise Exception(
                    f'keys to identify the location of VFile {f.path} '
                    f'contains an element not convertible to JSON: {key}'
                ) from e
            
        for struct_key, f in xfiles:
            if isinstance(f, IVFile):
                for k in struct_key:
                    _assert_key_json_convertible(k, f)

        # create method args
        method_args_ = [
            f.path if isinstance(f, IFile) else f for f in args_
        ]
        method_args, method_kwargs = \
            pack_sequence_as((args, kwargs), method_args_)

        # create Rule
        r = Rule(
            files_, xfiles, deplist,
            method, method_args, method_kwargs,
            force_update=force_update
        )

        # create RuleNode
        files = pack_sequence_as(files, files_)

        def conv_to_atom(x):
            assert isinstance(x, IFile)
            return FileNodeAtom(self._info, x)
            
        file_node_root = map_structure(
            conv_to_atom, files,
            seq_factory={list: FileNodeTuple, tuple: FileNodeTuple},
            map_factory={dict: FileNodeDict, Mapping: FileNodeDict}
        )

        fullname = (*self._name, name)
        if isinstance(file_node_root, FileNodeAtom):
            rc = RuleNodeAtom(fullname, r, self._info, file_node_root._file)
        elif isinstance(file_node_root, FileNodeTuple):
            rc = RuleNodeTuple(fullname, r, self._info, file_node_root)
        elif isinstance(file_node_root, FileNodeDict):
            rc = RuleNodeDict(fullname, r, self._info, file_node_root)

        # update group tree
        self._children[name] = rc

        if _ismembername(name):
            self.__dict__[name] = rc

        for f in files_:
            path_to_rule[f.abspath] = r

        for _k,f in xfiles:
            if f.abspath not in path_to_rule:
                path_to_rule[f.abspath] = None

        for f in itertools.chain(files_, (x[1] for x in xfiles)):
            path_to_file[f.abspath] = f

        self._info.rule_to_name[r] = '/'.join((*self._name, name))

        return rc


    def clean(self):
        for c in self._children.values():
            c.clean()

    def touch(self, _t=None):
        if _t is None:
            _t = time.time()
        for c in self._children.values():
            c.touch(_t)


    def _get_children_names(self, dst, group, rule):
        for k,c in self._children.items():
            if isinstance(c, Group):
                if group:
                    dst.append(c._name)

                c._get_children_names(dst, group, rule)
            else:
                if rule:
                    dst.append((*self._name, k))



    def select(self, pattern):
        if len(pattern) == 0:
            raise ValueError(f'Invalid pattern "{pattern}"')

        is_group = pattern[-1] == '/'

        pattern = pattern.strip('/')
        parts = re.split('/+', pattern)

        SEP = ';'
        regex = []
        for p in parts:
            assert len(p) > 0

            if p.find('**') != -1 and p != '**':
                raise ValueError(
                    'Invalid pattern: "**" can only be an entire component'
                )
            if p == '**':
                regex.append(f'({SEP}[^{SEP}]+)*')
            else:
                if p == '*':
                    # single * does not match an empty str
                    regex.append(f'{SEP}[^{SEP}]+')
                else:
                    def _repl(x):
                        x = x.group()
                        return f'[^{SEP}]*' if x == '*' else re.escape(x)
                    p = re.sub( r'\*|[^*]+', _repl, p)
                    regex.append(f'{SEP}{p}')

        regex = re.compile('^' + ''.join(regex) + '$')
        chnames = [self._name] if is_group else []
        self._get_children_names(chnames, is_group, not is_group)

        res = []
        for name in chnames:
            name = name[len(self._name):]
            if regex.match(''.join(SEP + n for n in name)):
                res.append(struct_get(self, name))
        
        return res


    def __repr__(self):
        name = repr_group_name(self._name)
        return f'<Group name={repr(name)} prefix={repr(self._prefix)}>'

    def __getitem__(self, k):
        return self._children[k]

    def __iter__(self):
        return iter(self._children)

    def __len__(self):
        return len(self._children)

    def __contains__(self, k):
        return k in self._children


def repr_group_name(group_name):
    return '/' + '/'.join(map(str, group_name))

def repr_rule_name(trg_name):
    return repr_group_name(trg_name[:-1]) + ':' + str(trg_name[-1])

def _ismembername(name):
    return \
        isinstance(name, str) and \
        name.isidentifier() and \
        name[0] != '_'


def _is_valid_node_name(name):
    # Group.select() depends on ':' being invalid
    return not re.search('[;*?"<>|]', name)


def make(
    *rule_or_groups,
    dry_run=False,
    keep_going=False,
    nthreads=1
):
    # create list of unique rules by DFS
    _added = set()
    rules = []
    stack = list(reversed(rule_or_groups))
    _info = None

    while stack:
        node = stack.pop()
        
        if _info is None:
            _info = node._info
        else:
            if _info is not node._info:
                raise ValueError(
                    f'All Groups/Rules must belong to the same Group tree. '
                    f'This rule is to prevent potentially erroneous operations'
                )

        if isinstance(node, RuleNodeBase):
            if node._rule not in _added:
                rules.append(node._rule)
        else:
            assert isinstance(node, Group)
            stack.extend(node._children.values())

    if nthreads <= 1:
        _make(rules, dry_run, keep_going, _info.callback)
    else:
        make_multi_thread(
            rules, dry_run, keep_going, nthreads, _info.callback)



def create_group(
    dirname=None, prefix=None, *,
    loglevel=None, use_default_logger=True, logfile=None):
    if (dirname is None) == (prefix is None):
        raise TypeError('Either dirname or prefix must be specified')

    if dirname is not None:
        assert isinstance(dirname, (str, os.PathLike))
        prefix = str(dirname) + os.path.sep

    assert isinstance(prefix, (str, os.PathLike))

    loglevel = loglevel or 'info'

    if logfile is None:
        logfiles = []
    elif isinstance(logfile, (list, tuple)):
        logfiles = logfile
    else:
        assert isinstance(logfile, (str, os.PathLike))
        logfiles = [logfile]

    logwriters = [_create_logwriter(f, loglevel) for f in logfiles]

    if use_default_logger:
        logwriters.append(_create_default_logwriter(loglevel))

    tree_info = GroupTreeInfo(logwriters=logwriters)

    return Group(tree_info, str(prefix), ())


def _create_default_logwriter(loglevel):
    if term_is_jupyter():
        return HTMLJupyterWriter(loglevel, os.getcwd())
    elif sys.stderr.isatty():
        return ColorTextWriter(sys.stderr, loglevel)
    else:
        return TextWriter(sys.stderr, loglevel)


def _create_logwriter(f, loglevel):
    if isinstance(f, (str, os.PathLike)):
        fname = str(Path(f))
        if fname[-5:] == '.html':
            return HTMLFileWriterOpenOnDemand(loglevel, fname)
        else:
            return TextFileWriterOpenOnDemand(loglevel, fname)
    else:
        if not (hasattr(f, 'write') and callable(f.write)):
            raise TypeError(f'{f} is not a writable stream')

        try:
            if f.isatty():
                return ColorTextWriter(f, loglevel)
        except:
            pass

        return TextWriter(f, loglevel)


SELF = StructKey(())



import sys, os, hashlib, json
from pathlib import Path, PurePath

from ..core.rule import Event, IRule
from .file import IFile, IVFile

class Rule(IRule):
    def __init__(
        self,
        yfiles,
        xfiles,
        deplist,
        method,
        args,
        kwargs,
        force_update=False
    ):
        self.yfiles = yfiles
        self.xfiles = xfiles
        self._deplist = deplist
        self._method = method
        self._args = args
        self._kwargs = kwargs
        self.force_update = force_update


    def should_update(
        self,
        updated_rules,
        dry_run
    ):
        for k,f in self.xfiles:
            if not os.path.exists(f.path):
                if dry_run:
                    return True
                else:
                    raise Exception(f'Input file {f.path} is missing')

            if os.path.getmtime(f.path) == 0:
                if dry_run:
                    return True
                else:
                    raise Exception(
                        f'Input file {f.path} has mtime of 0. '
                        f'Input files with mtime of 0 are considered to be '
                        f'invalid for reducing operational error.'
                    )

        if self.force_update:
            return True

        if dry_run and any(r in updated_rules for r in self._deplist):
            return True

        if any(not os.path.exists(f.path) for f in self.yfiles):
            return True

        oldest_y = min(os.path.getmtime(f.path) for f in self.yfiles)

        if oldest_y <= 0:
            return True

        xvfiles = [] # input VFiles that are updated

        for k,f in self.xfiles:
            if os.path.getmtime(f.path) > oldest_y:
                if isinstance(f, IVFile):
                    xvfiles.append((k,f))
                else:
                    return True

        if len(xvfiles) > 0:
            hash_dic = \
                {tuple(k): (h,t) for k,h,t in load_vfile_hashes(self.metadata_fname)}
                
            for k,f in xvfiles:
                if k not in hash_dic:
                    return True

                mtime = os.path.getmtime(f.path)
                hash_, mtime_ = hash_dic[k]

                # Optimization: skip computing hash if the current mtime
                # is equal to the one in the cache
                if mtime != mtime_ and f.get_hash() != hash_:
                    return True

        return False


    def preprocess(self, callback):
        for f in self.yfiles:
            try:
                os.makedirs(os.path.dirname(f.path), exist_ok=True)
            except:
                pass


    def postprocess(self, callback, succ):
        if succ:
            self.update_xvfile_hashes()
        else:
            # set mtime to 0
            for f in self.yfiles:
                try:
                    os.utime(f.path, (0, 0))
                except:
                    pass

            # delete vfile cache
            try:
                os.remove(self.metadata_fname)
            except:
                pass


    @property
    def metadata_fname(self):
        p = PurePath(self.yfiles[0].path)
        return p.parent / '.jtcmake' / p.name


    def update_xvfile_hashes(self):
        xvfiles = [(k,v) for k,v in self.xfiles if isinstance(v, IVFile)]
        if len(xvfiles) > 0:
            save_vfile_hashes(self.metadata_fname, xvfiles)


    @property
    def method(self): return self._method

    @property
    def args(self): return self._args

    @property
    def kwargs(self): return self._kwargs

    @property
    def deplist(self): return self._deplist


def create_vfile_hashes(vfiles):
    """
    list of (StructKey, file name, mtime)
    """
    res = [(k, f.get_hash(), os.path.getmtime(f.path)) for k,f in vfiles]
    res = json.loads(json.dumps(res)) # round trip JSON conversion
    return res
    

def save_vfile_hashes(metadata_fname, vfiles):
    hashes = create_vfile_hashes(vfiles)
    os.makedirs(os.path.dirname(metadata_fname), exist_ok=True)
    with open(metadata_fname, 'w') as f:
        try:
            json.dump(hashes, f)
        except Exception as e:
            raise ValueError(
                f'Failed to save IVFile hashes as JSON to {metadata_fname}.'
                f'This may be because some dictionary keys in the arguments'
                f' to specify the IVFile objects are not JSON convertible.'
            ) from e


def load_vfile_hashes(metadata_fname):
    if not os.path.exists(metadata_fname):
        return []

    with open(metadata_fname) as f:
        return json.load(f)





from collections.abc import Mapping, Sequence
import os, inspect
from pathlib import Path

from ..logwriter.writer import IWriter, RichStr
from ..core import events
from . import events as group_events


def create_event_callback(writers, rule_to_name):
    def callback(e):
        for w in writers:
            event_callback(w, rule_to_name, e)

    return callback


def event_callback(w, rule_to_name, e):
    if isinstance(e, events.ErrorRuleEvent):
        r = e.rule
        err = e.err
        name = rule_to_name.get(r)

        if name is None:
            w.warning(f'An event of unknown Rule has been emitted.\n')
            name = '<unknown>'

        if isinstance(e, events.UpdateCheckError):
            w.error(
                f'Failed to make {name}: '
                f'An error occured while checking if update is necessary: '
                f'{err}\n'
            )
        elif isinstance(e, events.PreProcError):
            w.error(
                f'Failed to make {name}: '
                f'An error occured during preprocessing: {err}\n'
            )
        elif isinstance(e, events.ExecError):
            w.error(
                f'Failed to make {name}: Method failed: {err}\n'
            )
        elif isinstance(e, events.PostProcError):
            w.error(
                f'Failed to make {name}: '
                f'An error occured during post-processing: {err}. '
                f'Make sure to remove the output files (if any) '
                f'by yourself\n'
            )
        elif isinstance(e, events.FatalError):
            w.error('Fatal error\n')
        else:
            w.warning(f'Unhandled error event for {r}: {err}\n')
        return
    elif isinstance(e, events.RuleEvent):
        r = e.rule
        name = rule_to_name.get(r)

        if name is None:
            w.warning(f'An event of unknown Rule has been emitted.\n')
            name = '<unknown>'

        if isinstance(e, events.Skip):
            msg = f'Skip {name}\n'
            if e.is_direct_target:
                w.info(msg)
            else:
                w.debug(msg)
        elif isinstance(e, events.Start):
            msg = []
            tostrs_func_call(msg, r.method, r.args, r.kwargs)
            msg = add_indent(msg, '  ')
            msg.insert(0, f'Make {name}\n')
            w.info(*msg)
        elif isinstance(e, events.Done):
            w.info(f'Done {name}\n')
        elif isinstance(e, events.DryRun):
            msg = []
            tostrs_func_call(msg, r.method, r.args, r.kwargs)
            msg = add_indent(msg, '  ')
            msg.insert(0, f'Make (dry) {name}\n')
            w.info(*msg)
        else:
            w.warning(f'Unhandled event for {r}\n')
        return
    elif isinstance(e, events.StopOnFail):
        w.warning(f'Execution aborted due to an error\n')
    elif isinstance(e, group_events.Clean):
        w.info('clean ', RichStr(str(e.path) + '\n', link=str(e.path)))
    elif isinstance(e, group_events.Touch):
        w.info('touch ', RichStr(str(e.path) + '\n', link=str(e.path)))
    else:
        w.warning(f'Unhandled event for {r}\n')
        


def add_indent(sl, indent):
    res = []
    for i,s in enumerate(sl):
        if i == 0 or (i > 0 and len(sl[i - 1]) > 0 and sl[i - 1][-1] == '\n'):
            res.append(indent + s)
        else:
            res.append(s)
    return res


def get_func_name(f):
    try:
        name, mod = f.__qualname__, f.__module__

        if mod == 'builtins' or mod == '__main__':
            return name
        else:
            return f'{mod}.{name}'
    except:
        return '<unkonw function>'


def tostrs_func_call(dst, f, args, kwargs):
    bn = inspect.signature(f).bind(*args, **kwargs)
    bn.apply_defaults()
    
    dst.append(get_func_name(f) + '(\n')

    for k,v in bn.arguments.items():
        dst.append(f'  {k}=')
        tostrs_obj(dst, v, capacity=500)
        dst.append(',\n')
    dst.append(')\n')


def tostrs_obj(dst, o, capacity=None):
    _tostrs_obj(dst, o, capacity or 10 ** 10)


def _tostrs_obj(dst, o, capacity):
    if isinstance(o, (tuple, list)):
        dst.append('[')
        for i,v in enumerate(o):
            if capacity <= 0:
                dst.append(', ...')
                break
            capacity = _tostrs_obj(dst, v, capacity)
            if i < len(o) - 1:
                dst.append(', ')
        dst.append(']')
    elif isinstance(o, Mapping):
        dst.append('{')
        for i,(k,v) in enumerate(o.items()):
            if capacity <= 0:
                dst.append(', ... ')
                break
            dst.append(repr(k) + ': ')
            capacity = _tostrs_obj(dst, v, capacity)
            if i < len(o) - 1:
                dst.append(', ')
        dst.append('}')
    elif isinstance(o, Path):
        res = RichStr(repr(o), link=str(o))
        dst.append(res)
        return capacity - len(res)
    else:
        res = repr(o)
        if len(res) > capacity:
            res = res[: capacity // 2] + ' ... ' + res[-capacity//2:]
        dst.append(res)
        return capacity - len(res)





import sys, html, abc, os

class RichStr(str):
    def __new__(cls, s, *_args, **_kwargs):
        return super().__new__(cls, s)


    def __init__(self, s, *, c=None, bg=None, link=None):
        if isinstance(s, RichStr):
            attr = s._attr.copy()
            a = {'c': c, 'bg': bg, 'link': link}
            attr.update({k:v for k,v in a.items() if v is not None})
        else:
            attr = {'c': c, 'bg': bg, 'link': link}

        # TODO: use frozen dict
        self._attr = attr

    
    def __getattr__(self, k):
        return self._attr[k]


    def __add__(self, rhs):
        if isinstance(rhs, type(self)):
            return NotImplemented # pass to __radd__
        elif isinstance(rhs, str):
            return RichStr(str(self) + str(rhs), self._attr)
        else:
            return NotImplemented


    def __radd__(self, lhs):
        if isinstance(lhs, str):
            return RichStr(str(lhs) + str(self), self._attr)
        else:
            return NotImplemented


    @property
    def attr(self):
        return self._attr


QUANT_LOG_LEVEL = {
    'debug': 10,
    'info': 20,
    'warning': 30,
    'error': 40,
}


class IWriter:
    def __init__(self, loglevel):
        assert loglevel in {'debug', 'info', 'warning', 'error'}

        self.loglevel = loglevel


    @abc.abstractmethod
    def _write(self, *args, level): ...

    def write(self, *args, level):
        if QUANT_LOG_LEVEL[self.loglevel] <= QUANT_LOG_LEVEL[level]:
            self._write(*args, level=level)

    def debug(self, *args):
        self.write(*args, level='debug')

    def info(self, *args):
        self.write(*args, level='info')

    def warning(self, *args):
        self.write(*args, level='warning')

    def error(self, *args):
        self.write(*args, level='error')

class TextWriter(IWriter):
    def __init__(self, writable, loglevel):
        super().__init__(loglevel)
        self.writable = writable


    def _write(self, *args, **kwargs):
        self.writable.write(''.join(map(str, args)))


class ColorTextWriter(IWriter):
    def __init__(self, writable, loglevel):
        super().__init__(loglevel)
        self.writable = writable


    def _write(self, *args, level):
        color = ({
            'debug': (0x4F, 0x8A, 0x10),
            'info': None,
            'warning': (0x9F, 0x60, 0x00),
            'error': (0xD8, 0x00, 0x0C),
        }).get(level)

        bgcolor = ({
            'debug': None,
            'info': None,
            'warning': None,
            'error': None,
        }).get(level)

        args = [RichStr(x, c=color, bg=bgcolor) for x in args]
        self.writable.write(create_color_str(args))


class TextFileWriterOpenOnDemand(IWriter):
    def __init__(self, loglevel, fname):
        super().__init__(loglevel)

        if not os.path.exists(os.path.dirname(fname)):
            raise FileNotFoundError(f'parent dir for {fname} not found')

        self.fname = fname

    def _write(self, *args, **kwargs):
        with open(self.fname, 'a') as f:
            f.write(''.join(map(str, args)))


HTML_BG_COLOR_MAP = {
    'debug': (0xDF, 0xF2, 0xBF),
    'info': (0xFF, 0xFF, 0xFF),
    'warning': (0xFE, 0xEF, 0xB3),
    'error': (0xFF, 0xD2, 0xD2),
}
HTML_COLOR_MAP = {
    'debug': (0x4F, 0x8A, 0x10),
    'info': (0, 0, 0),
    'warning': (0x9F, 0x60, 0x00),
    'error': (0xD8, 0x00, 0x0C),
}

class HTMLWriter(IWriter):
    def __init__(self, writable, loglevel, basedir=None):
        super().__init__(loglevel)
        self.writable = writable
        self._header = False
        self._footer = False
        self.basedir = basedir


    def _write(self, *args, level):
        color = HTML_COLOR_MAP.get(level, 'black')
        bgcolor = HTML_BG_COLOR_MAP.get(level, 'white')

        args = [RichStr(x, c=color, bg=bgcolor) for x in args]
        self.writable.write(create_html(args, self.basedir))


    def write_header(self):
        if self._header:
            return
        self._header = True
        self.writable.write(
            '<html><head><meta charset="utf-8"><title>log</title></head>'
            '\n<body><pre>'
        )

    def write_footer(self):
        if self._footer:
            return

        self._footer = True
        self.writable.write('</pre></body></html>')


    def __enter__(self):
        self.write_header()
        return self


    def __exit__(self, *args, **kwargs):
        self.write_footer()


class HTMLFileWriterOpenOnDemand(IWriter):
    def __init__(self, loglevel, fname, basedir=None):
        super().__init__(loglevel)

        if not os.path.exists(os.path.dirname(fname)):
            raise FileNotFoundError(f'parent dir for {fname} not found')

        self.basedir = basedir
        self.fname = fname

    def _write(self, *args, level):
        color = HTML_COLOR_MAP.get(level, 'black')
        bgcolor = HTML_BG_COLOR_MAP.get(level, 'white')

        args = [RichStr(x, c=color, bg=bgcolor) for x in args]

        with open(self.fname, 'a') as f:
            f.write(
                '<html><head><meta charset="utf-8"><title>log</title></head>'
                '<body><pre>'
            )
            f.write(create_html(args, self.basedir))
            f.write('</pre></body></html>')
        

class HTMLJupyterWriter(IWriter):
    def __init__(self, loglevel, basedir=None):
        super().__init__(loglevel)
        from IPython.display import display, HTML # check if importable
        self.basedir = basedir


    def _write(self, *args, level):
        from IPython.display import display, HTML
        color = HTML_COLOR_MAP.get(level, 'black')
        bgcolor = HTML_BG_COLOR_MAP.get(level, 'white')

        args = [RichStr(x, c=color, bg=bgcolor) for x in args]

        display(HTML(f'<pre>{create_html(args, self.basedir)}</pre>'))


def create_html(sl, basedir=None):
    sl = [x if isinstance(x, RichStr) else RichStr(x) for x in sl]
    groups = []
    for s in sl:
        if len(groups) >= 1 and s.attr == groups[-1][0].attr:
            groups[-1].append(s)
        else:
            groups.append([s])

    outs = []
    for group in groups:
        s = RichStr(''.join(group), **group[0].attr)
        outs.append(_richstr_to_html(s, basedir))

    return ''.join(outs)


def _richstr_to_html(s, basedir):
    starts = []
    ends = []

    if s.link is not None:
        if basedir is not None:
            rel = os.path.relpath(s.link, basedir)
            starts.append(f'<a href="{rel}">')
        else:
            starts.append(f'<a href="{s.link}">')
        ends.append(f'</a>')

    styles = []
    if s.c is not None:
        styles.append(f'color: rgb({s.c[0]}, {s.c[1]}, {s.c[2]});')

    if s.bg is not None:
        styles.append(
            f'background-color: rgb({s.bg[0]}, {s.bg[1]}, {s.bg[2]});')

    if len(styles) != 0:
        style = ''.join(styles)
        starts.append(f'<span style="{style}">')
        ends.append(f'</span>')

    starts.append(html.escape(s))
    starts.extend(ends[::-1])
    return ''.join(starts)
    

def create_color_str(sl):
    res = []
    last_c = None
    last_bg = None
    for s in sl:
        if s.bg != last_bg:
            last_bg = s.bg
            res.append(f'\x1b[48;5;{_comp_8bit_term_color(*s.bg)}m')

        if s.c != last_c:
            last_c = s.c
            res.append(f'\x1b[38;5;{_comp_8bit_term_color(*s.c)}m')

        res.append(str(s))

    return ''.join(res)


def _comp_8bit_term_color(r, g, b):
    return 16 + r * 36 + g * 6 + b


def term_is_jupyter():
    try:
        from IPython.core.getipython import get_ipython
        return get_ipython().__class__.__name__ == 'ZMQInteractiveShell'
    except:
        return False



import sys, os, html, re
from abc import ABCMeta, abstractmethod
from .simplehtmlbuilder import Element, to_html_str
from ..utils import isipynb

class Writer:
    def __init__(self, writable):
        self.writable = writable

    def __call__(self, text, link_map=None, logkind=None):
        self.writable.write(text)


class HTMLWriter:
    def __init__(self, writable):
        self.f = writable
        self.state = 0

    def __enter__(self):
        if self.state != 0:
            raise ValueError('Context may not be entered multiple times')
        self.state = 1

        self.f.write(
            '''
            <html>
                <head><meta charset="utf-8"><title>log</title></head>
                <body>
            '''
        )

        return self

    def __exit__(self, *args, **kwargs):
        if self.state == 1:
            self.state = 2
            self.f.write('</body></html>')

    def __call__(self, obj, link_map=None, logkind='info'):
        """logkind is log|info|warning|error|success"""
        if self.state != 1:
            raise ValueError('this must be used under with')

        self.f.write(to_html_str(create_msg_html(obj, link_map, logkind)))


class HTMLIpynbWriter:
    def __call__(self, obj, link_map=None, logkind='info'):
        from IPython.display import display, HTML
        obj = create_msg_html(obj, link_map, logkind)
        display(HTML(to_html_str(obj)))


def create_msg_html(obj, link_map=None, logkind='info'):
    if link_map is not None:
        obj = replace_link(obj, link_map)

    bgcolormap = {
        'success': '#DFF2BF',
        'log': 'white',
        'warning': '#FEEFB3',
        'error': '#FFD2D2',
    }
    colormap = {
        'success': '#4F8A10',
        'log': 'grey',
        'warning': '#9F6000',
        'error': '#D8000C',
    }
    color = colormap.get(logkind, 'black')
    bgcolor = bgcolormap.get(logkind, 'white')

    style = f'color: {color}; background-color: {bgcolor}'

    obj = Element('pre', Element('span', obj, style=style))

    return obj


def replace_link_of_str(s, link_map):
    if len(link_map) == 0:
        return s
    p = '|'.join(re.escape(l) for l in link_map.keys())
    p = rf'([\s\S]*?)({p})|([\s\S]+)'

    res = []
    for m in re.finditer(p, s):
        a,l,tail = m.groups()
        if tail is None:
            res.append(a)
            res.append(Element('a', l, href=link_map[l]))
        else:
            res.append(tail)
            
    return res

def replace_link(obj, link_map):
    if isinstance(obj, (tuple, list)):
        res = []
        return [replace_link(v, link_map) for v in obj]
    elif isinstance(obj, str):
        return replace_link_of_str(obj, link_map)
    elif isinstance(obj, Element):
        if obj.tag == 'a':
            return obj
        c = replace_link(obj.children, link_map)
        return Element(obj.tag, c, **obj.kwargs)


class TermWriter:
    def __init__(self, out):
        self.out = out

    def __call__(self, msg, link_map=None, logkind='info'):
        colormap = {
            'success': '36',
            'log': '37',
            'warning': '33',
            'error': '31',
        }
        color = colormap.get(logkind, '0')

        self.out.write(f'\x1b[{color}m{msg}\x1b[0m')
        

def get_default_writer():
    if isipynb():
        return HTMLIpynbWriter()
    else:
        if sys.stderr.isatty():
            return TermWriter(sys.stderr)
        else:
            return Writer(sys.stderr)




import os, sys, re, traceback, inspect, abc, time, enum
from threading import Condition, Thread
from collections import defaultdict, deque

from . import events
from .rule import Event, IRule

class Result(enum.Enum):
    Update = 1
    Skip = 2
    Fail = 3


def make(
    rules,
    dry_run,
    keep_going,
    callback,
    ):
    if len(rules) == 0:
        return

    direct_targets = set(rules)

    added = set()
    taskq = []

    # topological sort
    def rec(t):
        if t in added:
            return

        added.add(t)

        for dept in t.deplist:
            rec(dept)

        taskq.append(t)

    for rule in rules:
        rec(rule)

    failed_rule = set()
    updated_rules = set()

    for t in taskq:
        if any(dept in failed_rule for dept in t.deplist):
            failed_rule.add(t)
            continue

        try:
            result = process_rule(
                t, dry_run, updated_rules, direct_targets, callback
            )
        except Exception as e:
            traceback.print_exc()
            callback(events.FatalError(t, e))
            return False

        if result == Result.Update:
            updated_rules.add(t)
        elif result == Result.Fail:
            failed_rule.add(t)
            if not keep_going:
                callback(events.StopOnFail())
                return False

    return True
        

def process_rule(
    rule,
    dry_run,
    updated_rules,
    direct_targets,
    callback
    ):
    if dry_run:
        try:
            should_update = rule.should_update(updated_rules, True)
        except Exception as e:
            traceback.print_exc()
            callback(events.UpdateCheckError(rule, e))
            return Result.Fail

        if should_update:
            callback(events.DryRun(rule))
            return Result.Update
        else:
            callback(events.Skip(rule))
            return Result.Skip

    try:
        should_update = rule.should_update(updated_rules, False)
    except Exception as e:
        traceback.print_exc()
        callback(events.UpdateCheckError(rule, e))
        return Result.Fail

    if not should_update:
        callback(events.Skip(rule, rule in direct_targets))
        return Result.Skip

    callback(events.Start(rule))

    try:
        rule.preprocess(callback)
    except Exception as e:
        callback(events.PreProcError(rule, e))
        return Result.Fail

    try:
        rule.method(*rule.args, **rule.kwargs)
        succ = True
    except Exception as e:
        traceback.print_exc()
        callback(events.ExecError(rule, e))
        succ = False

    try:
        rule.postprocess(callback, succ)
    except Exception as e:
        callback(events.PostProcError(rule, e))
        return Result.Fail

    if succ:
        callback(events.Done(rule))
        return Result.Update
    else:
        return Result.Fail


def make_multi_thread(
    rules,
    dry_run,
    keep_going,
    nthreads,
    callback
    ):
    if nthreads < 1:
        raise ValueError('nthreads must be greater than 0')

    if len(rules) == 0:
        return

    direct_targets = set(rules)

    b2a = defaultdict(set) # before to after
    dep_cnt = {}

    def rec(t):
        if t in dep_cnt:
            return

        dep_cnt[t] = len(t.deplist)

        for dept in t.deplist:
            b2a[dept].add(t)
            rec(dept)
    
    for t in rules:
        rec(t)

    updated_rules = set()

    cv = Condition()
    taskq = deque(t for t,c in dep_cnt.items() if c == 0)
    processing = set()
    stop = False

    def get_rule_fn():
        with cv:
            while len(taskq) == 0 and len(processing) != 0 and not stop:
                cv.wait()

            if stop:
                return None

            if len(taskq) == 0 and len(processing) == 0:
                return None

            r = taskq.popleft()
            processing.add(r)

            return r

    def set_result_fn(rule, res):
        nonlocal stop
        with cv:
            processing.remove(rule)

            if res is None:
                stop = True
            elif res == Result.Fail:
                if not keep_going:
                    stop = True
            else:
                if res == Result.Update:
                    updated_rules.add(res)

                for nxt in b2a[rule]:
                    dep_cnt[nxt] -= 1
                    if dep_cnt[nxt] == 0:
                        taskq.append(nxt)

            cv.notify_all()
            

    args = (
        get_rule_fn, set_result_fn,
        updated_rules, direct_targets, dry_run, callback)
    threads = [
        Thread(target=worker, args=(*args,), name=f'lightmake{i}')
        for i in range(nthreads)
    ]

    for t in threads: t.start()
    for t in threads: t.join()

    if stop:
        callback(events.StopOnFail(None))


def worker(
    get_rule_fn,
    set_result_fn,
    updated_rules,
    direct_targets,
    dry_run,
    callback,
    ):
    while True:
        rule = get_rule_fn()

        if rule is None:
            return
        
        res = None

        try:
            res = process_rule(
                rule, dry_run, updated_rules, direct_targets, callback
            )
        except Exception as e:
            traceback.print_exc()
            callback(events.FatalError(rule, e))
        finally:
            set_result_fn(rule, res)




import os, sys, re, json, subprocess
from html import escape

from ..logwriter.writer import term_is_jupyter
from .igroup import IGroup


def print_graphviz(group, output_file=None):
    if output_file is None:
        if term_is_jupyter():
            from IPython.display import display, SVG
            dot_code = gen_dot_code(group)
            svg = convert(dot_code, 'svg').decode()
            display(SVG(svg))
            return
        else:
            print(gen_dot_code(group))
            return
    else:
        dot_code = \
            gen_dot_code(group, os.path.dirname(output_file))

        if output_file[-4:] == '.svg':
            data = convert(dot_code, 'svg')
        else:
            raise ValueError(f'Output file\'s extension must be .svg')

        with open(output_file, 'wb') as f:
            f.write(data)


def gen_dot_code(group, basedir=None):
    if not isinstance(group, IGroup):
        raise TypeError('argument group must be Group')

    gid = {}
    rid = {}
    fid = {}

    res = []
    res.append('digraph {')
    res.append('  compound=true;')

    def rec_group(g, idt, par_prefix):
        gid[g] = len(gid)

        name = '<ROOT>' if len(g._name) == 0 else g._name[-1]

        if par_prefix == '':
            prefix = g._prefix
        elif g._prefix[:len(par_prefix)] == par_prefix:
            prefix = '... ' + g._prefix[len(par_prefix):]
        else:
            prefix = g._prefix

        res.append(idt + f'subgraph cluster{gid[g]} {{')
        res.append(
            idt +
            f'  label = <<B>{escape(name)}</B> '
            f'(<FONT FACE="monospace">{escape(prefix)}</FONT>)>;')
        res.append(idt + f'  style = "rounded";')

        for cname in g:
            c = g[cname]
            if isinstance(c, IGroup):
                rec_group(c, idt+'  ', g._prefix)
            else:
                proc_rulew(c, cname, idt + '  ', g._prefix)

        res.append(idt + '};');

    def proc_rulew(rw, name, idt, par_prefix):
        r = rw._rule
        rid[r] = len(rid)

        res.append(idt + f'subgraph cluster_r_{rid[r]} {{')
        res.append(idt + f'  label=<<B>{escape(name)}</B>>;')
        res.append(idt + f'  bgcolor = "#E0FFFF";')

        par_prefix = os.path.abspath(par_prefix + '_')[:-1]

        for yf in r.yfiles:
            fid[yf] = len(fid)

            p = str(yf.abspath)
            if par_prefix != '' and p[:len(par_prefix)] == par_prefix:
                p = '... ' + p[len(par_prefix):]
            else:
                p = str(yf.path)

            res.append(
                idt +
                f'  f{fid[yf]} ['
                f'label=<<FONT FACE="monospace">{escape(p)}</FONT>>; '
                f'style=filled; '
                f'color=white; '
                f'shape=plain; '
                f'margin="0.1,0.1"; '
                f'URL="{_mk_link(yf.path, basedir)}"; '
                f'];'
            )
        res.append(idt + f'}}')

    rec_group(group, '  ', '')

    for r, i in rid.items():
        for _k,xf in r.xfiles:
            if xf not in fid:
                fid[xf] = len(fid)
                res.append(
                    f'  f{fid[xf]} ['
                    f'label=<<FONT FACE="monospace">'
                    f'{escape(str(xf.path))}</FONT>>; '
                    f'shape=plain; '
                    f'URL="{_mk_link(xf.path, basedir)}"; '
                    f'];'
                )


            res.append(
                f'  f{fid[xf]} -> f{fid[r.yfiles[0]]} '
                f'[lhead=cluster_r_{rid[r]}];'
            )

    res.append('}')

    return '\n'.join(res) + '\n'


def convert(dot_code, t='svg'):
    p = subprocess.run(
        ['dot', f'-T{t}'],
        input=dot_code.encode(),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    if p.returncode != 0:
        sys.stderr.write(p.stderr.decode())
        raise Exception(f'Error: dot exit with code {p.returncode}')

    return p.stdout


def save_to_file(dot_code, fname, t='svg'):
    with open(fname, 'wb') as f:
        f.write(convert(dot_code, t))


def _mk_link(p, basedir):
    basedir = basedir or os.getcwd()

    try:
        return os.path.relpath(p, basedir)
    except:
        pass

    return str(p)



from .rule import Event, IRule

class RuleEvent(Event):
    def __init__(self, rule):
        super().__init__()
        self._rule = rule

    @property
    def rule(self):
        return self._rule


    def __repr__(self):
        return f'{type(self).__name__}({self.rule})'


class ErrorRuleEvent(RuleEvent):
    def __init__(self, rule, err: Exception):
        super().__init__(rule)
        self._error = err

    @property
    def err(self):
        return self._error


    def __repr__(self):
        return f'{type(self).__name__}(rule={self.rule}, err={self.err})'


class Skip(RuleEvent):
    def __init__(self, rule, is_direct_target):
        super().__init__(rule)
        self.is_direct_target = is_direct_target

class Start(RuleEvent): ...

class Done(RuleEvent): ...

class DryRun(RuleEvent): ...

class StopOnFail(Event): ...

class UpdateCheckError(ErrorRuleEvent): ...

class PreProcError(ErrorRuleEvent): ...

class ExecError(ErrorRuleEvent): ...

class PostProcError(ErrorRuleEvent): ...

class FatalError(ErrorRuleEvent): ...




import abc

class Event:
    def __init__(self, msg=None):
        self.msg = msg


class IRule(abc.ABC):
    @abc.abstractmethod
    def should_update(self, updated_rules, dry_run): ...

    @abc.abstractmethod
    def preprocess(self, callback): ...

    @abc.abstractmethod
    def postprocess(self, callback, succ): ...

    @property
    @abc.abstractmethod
    def method(self): ...

    @property
    @abc.abstractmethod
    def args(self): ...

    @property
    @abc.abstractmethod
    def kwargs(self): ...

    @property
    @abc.abstractmethod
    def deplist(self): ...




import sys, os, html, re

_VOIDTAG = {'area', 'base', 'br', 'col', 'embed', 'hr', 'img', 'input', 'link', 'meta', 'source', 'track', 'wbr'}

class Element:
    def __init__(self, tag, children, **kwargs):
        self.tag = tag.lower()
        self.kwargs = kwargs
        self.children = children

    def __str__(self):
        return to_html_str(self)


class EmptyElement(Element):
    def __init__(self, tag, **kwargs):
        super().__init__(tag, [], **kwargs)

    def __call__(self, *args):
        return Element(self.tag, args, **self.kwargs)
        


def to_html_str(obj):
    lst = []

    def rec(obj):
        if isinstance(obj, (tuple, list)):
            for c in obj:
                rec(c)
        elif isinstance(obj, str):
            lst.append(html.escape(obj))
        elif isinstance(obj, Element):
            kv2prop = lambda kv: f' {kv[0]}="{html.escape(str(kv[1]))}"'
            prop = ''.join(map(kv2prop, obj.kwargs.items()))

            if obj.tag in _VOIDTAG:
                lst.append(f'<{obj.tag}{prop}/>')
            else:
                lst.append(f'<{obj.tag}{prop}>')
                rec(obj.children)
                lst.append(f'</{obj.tag}>')
        elif obj is None:
            pass
        else:
            lst.append(html.escape(str(obj)))

    rec(obj)

    return ''.join(lst)
    



from collections.abc import Mapping


class StructKey(tuple):
    def __new__(cls, struct_key):
        return super().__new__(cls, struct_key)

    def __getitem__(self, key):
        if not isinstance(key, (int, str)):
            raise ValueError(f'Key must be int or str. Given {key}')

        return StructKey((*self, key))

    def __getattr__(self, key):
        return self[key]

    def __repr__(self):
        return f'StructKey({super().__repr__()})'


def struct_get(struct, struct_key):
    for k in struct_key:
        struct = struct[k]

    return struct


def map_structure(
    map_fn, struct,
    seq_factory={list: list, tuple: tuple},
    map_factory={(dict, Mapping): dict}
):
    assert callable(map_fn)

    def rec(struct):
        if isinstance(struct, StructKey):
            return map_fn(struct)

        for src, dst in seq_factory.items():
            if isinstance(struct, src):
                return dst(map(rec, struct))

        for src, dst in map_factory.items():
            if isinstance(struct, src):
                return dst({k: rec(struct[k]) for k in struct.keys()})

        return map_fn(struct)

    return rec(struct)


def ordered_map_structure(
    map_fn, struct,
    seq_factory={list: list, tuple: tuple},
    map_factory={(dict, Mapping): dict}
):
    assert callable(map_fn)

    def rec(struct):
        if isinstance(struct, StructKey):
            return map_fn(struct)

        for src, dst in seq_factory.items():
            if isinstance(struct, src):
                return dst(map(rec, struct))

        for src, dst in map_factory.items():
            if isinstance(struct, src):
                keys = sorted(struct.keys(), key=lambda x: (hash(x), x))
                return dst({k: rec(struct[k]) for k in keys})

        return map_fn(struct)

    return rec(struct)



def flatten(struct):
    res = []

    def rec(node):
        if isinstance(node, StructKey):
            res.append(node)
        elif isinstance(node, (tuple, list)):
            for v in node:
                rec(v)
        elif isinstance(node, (dict, Mapping)):
            keys = sorted(node.keys(), key=lambda x: (hash(x), x))
            for k in keys:
                rec(node[k])
        else:
            res.append(node)

    rec(struct)
    return res


def flatten_to_struct_keys(struct):
    res = []

    def rec(node, struct_key):
        if isinstance(node, StructKey):
            res.append(struct_key)
        elif isinstance(node, (tuple, list)):
            for i,v in enumerate(node):
                rec(v, (*struct_key, i))
        elif isinstance(node, (dict, Mapping)):
            keys = sorted(node.keys(), key=lambda x: (hash(x), x))
            for k in keys:
                rec(node[k], (*struct_key, k))
        else:
            res.append(struct_key)

    rec(struct, ())
    return res


class NotEnoughElementError(Exception):
    def __init__(self, msg=None):
        super().__init__(msg or 'not enough elements in flatten seq')


def pack_sequence_as(ref_struct, flatten_seq):
    i = 0

    def map_fn(x):
        nonlocal i
        i += 1
        if i > len(flatten_seq):
            raise NotEnoughElementError()
        return flatten_seq[i - 1]

    res, err = None, None

    try:
        res = ordered_map_structure(map_fn, ref_struct)
    except NotEnoughElementError as e:
        err = e

    if err is not None:
        raise err

    if i != len(flatten_seq):
        raise TypeError('too many elements in flatten_seq')

    return res



import sys, os
from ..core.rule import IRule
from ..logwriter.writer import term_is_jupyter, create_html
from .event_logger import tostrs_func_call


def print_method(rule):
    rule = rule._rule
    sl = []
    tostrs_func_call(sl, rule.method, rule.args, rule.kwargs)

    if term_is_jupyter():
        from IPython.display import display, HTML
        display(HTML('<pre>' + create_html(sl, os.getcwd()) + '</pre>'))
    else:
        sys.stdout.write(''.join(sl))



from abc import ABC, abstractmethod
from collections.abc import Mapping

class IGroup(ABC):
    @abstractmethod
    def add_group(self, name, dirname, *, prefix): ...

    @abstractmethod
    def add(self, name, *args, **kwargs): ...

    @abstractmethod
    def addvf(self, name, *args, **kwargs): ...

    @abstractmethod
    def make(self, dry_run=False, keep_going=False, *, nthreads=1): ...

    @abstractmethod
    def clean(self): ...

    @abstractmethod
    def touch(self, _t=None): ...
    
    @abstractmethod
    def select(self, pattern): ...

    @abstractmethod
    def __getitem__(self, k): ...

    @abstractmethod
    def __iter__(self): ...

    @abstractmethod
    def __len__(self): ...

    @abstractmethod
    def __contains__(self, k): ...




from abc import ABC, abstractmethod
import os, time, hashlib, base64
from pathlib import PurePath, Path

from ..core.rule import IRule

class IFile(ABC):
    @property
    @abstractmethod
    def path(self): ...

    @property
    @abstractmethod
    def abspath(self): ...


class IVFile(IFile):
    @abstractmethod
    def get_hash(self): ...


class File(IFile):
    def __init__(self, path):
        assert isinstance(path, (str, os.PathLike))
        self._path = Path(path)

    @property
    def path(self):
        return self._path

    @property
    def abspath(self):
        return self._path.absolute()

    def __hash__(self):
        return hash(self._path)

    def __eq__(self, other):
        return type(other) == type(self) and self._path == other._path

    def __repr__(self):
        return f'{type(self).__name__}(path={repr(self.path)})'


class VFile(IVFile, File):
    def _clean(self):
        try:
            os.remove(self._path)
            os.remove(self.metadata_fname)
        except:
            pass

    def get_hash(self):
        return get_hash(self.path)
        

_hash_cache = {}

def get_hash(fname):
    fname = os.path.realpath(fname)

    if fname in _hash_cache:
        if os.path.getmtime(fname) == _hash_cache[fname][0]:
            return _hash_cache[fname][1]

    mtime = os.path.getmtime(fname)

    with open(fname, 'rb') as f:
        res = base64.b64encode(hashlib.md5(f.read()).digest()).decode()

    _hash_cache[fname] = (mtime, res)

    return res
